@@ -33,6 +33,7 @@ from ._hdbscan_boruvka import KDTreeBoruvkaAlgorithm, BallTreeBoruvkaAlgorithm
 from .dist_metrics import DistanceMetric
 
 from .plots import CondensedTree, SingleLinkageTree, MinimumSpanningTree
+from .prediction import PredictionData
 
 FAST_METRICS = (KDTree.valid_metrics + BallTree.valid_metrics +
                 ['cosine', 'arccos'])
@@ -93,9 +94,10 @@ else:
 
 
 def _tree_to_labels(X, single_linkage_tree, min_cluster_size=10,
+                    cluster_selection_method='eom',
                     allow_single_cluster=False,
                     match_reference_implementation=False):
-    """ Converts a pretrained tree and cluster size into a
+    """Converts a pretrained tree and cluster size into a
     set of labels and probabilities.
     """
     condensed_tree = condense_tree(single_linkage_tree,
@@ -103,6 +105,7 @@ def _tree_to_labels(X, single_linkage_tree, min_cluster_size=10,
     stability_dict = compute_stability(condensed_tree)
     labels, probabilities, stabilities = get_clusters(condensed_tree,
                                                       stability_dict,
+                                                      cluster_selection_method,
                                                       allow_single_cluster,
                                                       match_reference_implementation)
 
@@ -331,7 +334,8 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
             metric='minkowski', p=2, leaf_size=40,
             algorithm='best', memory=Memory(cachedir=None, verbose=0),
             approx_min_span_tree=True, gen_min_span_tree=False,
-            core_dist_n_jobs=4, allow_single_cluster=False,
+            core_dist_n_jobs=4,
+            cluster_selection_method='eom', allow_single_cluster=False,
             match_reference_implementation=False, **kwargs):
     """Perform HDBSCAN clustering from a vector array or distance matrix.
 
@@ -404,6 +408,14 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
         supported by the specific algorithm). For ``core_dist_n_jobs``
         below -1, (n_cpus + 1 + core_dist_n_jobs) are used.
 
+    cluster_selection_method : string, optional (default='eom')
+        The method used to select clusters from the condensed tree. The
+        standard approach for HDBSCAN* is to use an Excess of Mass algorithm
+        to find the most persistent clusters. Alternatively you can instead
+        select the clusters at the leaves of the tree -- this provides the
+        most fine grained and homogeneous clusters. Options are:
+            * ``eom``
+            * ``leaf``
 
     allow_single_cluster : bool, optional (default=False)
         By default HDBSCAN* will not produce a single cluster, setting this
@@ -495,6 +507,10 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
         min_cluster_size = min_cluster_size + 1
         approx_min_span_tree = False
 
+    if cluster_selection_method not in ('eom', 'leaf'):
+        raise ValueError('Invalid Cluster Selection Method: %s\n'
+                         'Should be one of: "eom", "leaf"\n')
+
     # Checks input and converts to an nd-array where possible
     X = check_array(X, accept_sparse='csr')
     # Python 2 and 3 compliant string_type checking
@@ -592,6 +608,7 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
     return _tree_to_labels(X,
                            single_linkage_tree,
                            min_cluster_size,
+                           cluster_selection_method,
                            allow_single_cluster,
                            match_reference_implementation) + \
             (result_min_span_tree,)
@@ -673,11 +690,27 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
         supported by the specific algorithm). For ``core_dist_n_jobs``
         below -1, (n_cpus + 1 + core_dist_n_jobs) are used.
 
+    cluster_selection_method : string, optional (default='eom')
+        The method used to select clusters from the condensed tree. The
+        standard approach for HDBSCAN* is to use an Excess of Mass algorithm
+        to find the most persistent clusters. Alternatively you can instead
+        select the clusters at the leaves of the tree -- this provides the
+        most fine grained and homogeneous clusters. Options are:
+            * ``eom``
+            * ``leaf``
+
     allow_single_cluster : bool, optional (default=False)
         By default HDBSCAN* will not produce a single cluster, setting this
         to True will override this and allow single cluster results in
         the case that you feel this is a valid result for your dataset.
 
+    prediction_data : boolean, optional
+        Whether to generate extra cached data for predicting labels or
+        membership vectors few new unseen points later. If you wish to
+        persist the clustering object for later re-use you probably want
+        to set this to True.
+        (default False)
+
     match_reference_implementation : bool, optional (default=False)
         There exist some interpretational differences between this
         HDBSCAN* implementation and the original authors reference
@@ -727,6 +760,11 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
         outlier-like the point. Useful as an outlier detection technique.
         Based on the GLOSH algorithm by Campello, Moulavi, Zimek and Sander.
 
+    prediction_data_ : PredictionData object
+        Cached data used for predicting the cluster labels of new or
+        unseen points. Necessary only if you are using functions from
+        ``hdbscan.prediction``.
+
     References
     ----------
 
@@ -753,7 +791,9 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
                  approx_min_span_tree=True,
                  gen_min_span_tree=False,
                  core_dist_n_jobs=4,
+                 cluster_selection_method='eom',
                  allow_single_cluster=False,
+                 prediction_data=False,
                  match_reference_implementation=False, **kwargs):
         self.min_cluster_size = min_cluster_size
         self.min_samples = min_samples
@@ -767,8 +807,10 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
         self.approx_min_span_tree = approx_min_span_tree
         self.gen_min_span_tree = gen_min_span_tree
         self.core_dist_n_jobs = core_dist_n_jobs
+        self.cluster_selection_method = cluster_selection_method
         self.allow_single_cluster = allow_single_cluster
         self.match_reference_implementation = match_reference_implementation
+        self.prediction_data = prediction_data
 
         self._metric_kwargs = kwargs
 
@@ -777,6 +819,7 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
         self._min_spanning_tree = None
         self._raw_data = None
         self._outlier_scores = None
+        self._prediction_data = None
 
     def fit(self, X, y=None):
         """Perform HDBSCAN clustering from features or distance matrix.
@@ -798,6 +841,9 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
             self._raw_data = X
 
         kwargs = self.get_params()
+        # prediction data only applies to the persistent model, so remove
+        # it from the keyword args we pass on the the function
+        kwargs.pop('prediction_data', None)
         kwargs.update(self._metric_kwargs)
 
         (self.labels_,
@@ -807,6 +853,9 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
          self._single_linkage_tree,
          self._min_spanning_tree) = hdbscan(X, **kwargs)
 
+        if self.prediction_data:
+            self.generate_prediction_data()
+
         return self
 
     def fit_predict(self, X, y=None):
@@ -827,6 +876,41 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
         self.fit(X)
         return self.labels_
 
+    def generate_prediction_data(self):
+        """
+        Create data that caches intermediate results used for predicting
+        the label of new/unseen points. This data is only useful if
+        you are intending to use functions from ``hdbscan.prediction``.
+        """
+
+        if self.metric in FAST_METRICS:
+            min_samples = self.min_samples or self.min_cluster_size
+            if self.metric in KDTree.valid_metrics:
+                tree_type = 'kdtree'
+            elif self.metric in BallTree.valid_metrics:
+                tree_type = 'balltree'
+            else:
+                warn('Metric {} not supported for prediction data!'.format(self.metric))
+                return
+                
+            self._prediction_data = PredictionData(
+                self._raw_data, self.condensed_tree_, min_samples,
+                tree_type=tree_type, metric=self.metric,
+                **self._metric_kwargs
+            )
+        else:
+            warn('Cannot generate prediction data for non-vector'
+                 'space inputs -- access to the source data rather'
+                 'than mere distances is required!')
+
+    @property
+    def prediction_data_(self):
+        if self._prediction_data is None:
+            warn('No prediction data was generated')
+            return None
+        else:
+            return self._prediction_data
+
     @property
     def outlier_scores_(self):
         if self._outlier_scores is not None:
@@ -842,7 +926,7 @@ class HDBSCAN(BaseEstimator, ClusterMixin):
     @property
     def condensed_tree_(self):
         if self._condensed_tree is not None:
-            return CondensedTree(self._condensed_tree)
+            return CondensedTree(self._condensed_tree, self.cluster_selection_method)
         else:
             warn('No condensed tree was generated; try running fit first.')
             return None
